﻿/*
 * wg12232.c
 * Модуль для работы с дисплеем WG12232A, или совместимым
 * 
 * Author: Погребняк Дмитрий, г. Самара, 2013
 *
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here are free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 

#include "wg12232.h"
#include <avr/io.h>
#include <util/delay.h>
#include "Tools.h"


uint8_t wg12232_chip_page;
 
inline void wg12232_strobe_on(uint8_t chip) {
  if (chip & 1)
    WG12232_CS_PORT |= (1 << WG12232_CS1_PIN);
  if (chip & 2)
    WG12232_CS_PORT |= (1 << WG12232_CS2_PIN);
}

inline void wg12232_strobe_off(uint8_t chip) {
  WG12232_CS_PORT &= ~((1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN));
}

inline void wg12232_set_data_port_direction(uint8_t direction) {
#ifdef WG12232_DATA_PORT_DIRECTION 
  WG12232_DATA_PORT_DIRECTION = direction;
#else
  // Let's optimizer to select one from below
  // Оптимизатор оставит только один из вариантов
  if (&WG12232_DATA_PORT == &PORTA) {
    DDRA = direction;
  } else if (&WG12232_DATA_PORT == &PORTB) {
    DDRB = direction;
  } else if (&WG12232_DATA_PORT == &PORTC) {
    DDRC = direction;
  } else if (&WG12232_DATA_PORT == &PORTD) {
    DDRD = direction;
  }
#endif
}

inline uint8_t wg12232_read_data_port_pins() {
#ifdef WG12232_DATA_PORT_PINS
  return WG12232_DATA_PORT_PINS;
#else
  // Let's optimizer to select one from below
  // Оптимизатор оставит только один из вариантов
  if (&WG12232_DATA_PORT == &PORTA) {
    return PINA;
  } else if (&WG12232_DATA_PORT == &PORTB) {
    return PINB;
  } else if (&WG12232_DATA_PORT == &PORTC) {
    return PINC;
  } else  {
    return PIND;
  }
#endif
}

inline void wg12232_delay() {
  #if defined(F_CPU) && (F_CPU > 10000000)
    asm volatile ("nop");
    #if (F_CPU > 18000000) 
      asm volatile ("nop");
    #endif
  #endif
  asm volatile ("nop");
}

void wg12232_command(uint8_t command, uint8_t chip) {
  wg12232_set_data_port_direction(0xFF);
  WG12232_A0RW_PORT &= ~(1 << WG12232_RW_PIN);
  WG12232_A0RW_PORT &= ~(1 << WG12232_A0_PIN);
  WG12232_DATA_PORT = command;
  wg12232_strobe_on(chip);
  wg12232_delay();
  wg12232_strobe_off(chip);
  WG12232_DATA_PORT = 0x00;
  wg12232_set_data_port_direction(0x00);
}

void wg12232_write(uint8_t data, uint8_t chip) {
  wg12232_set_data_port_direction(0xFF);
  WG12232_A0RW_PORT &= ~(1 << WG12232_RW_PIN);
  WG12232_A0RW_PORT |= (1 << WG12232_A0_PIN);
  WG12232_DATA_PORT = data;
  wg12232_strobe_on(chip);
  wg12232_delay();
  wg12232_strobe_off(chip);
  WG12232_A0RW_PORT &= ~(1 << WG12232_A0_PIN);
  WG12232_DATA_PORT = 0x00;
  wg12232_set_data_port_direction(0x00);
}

uint8_t wg12232_status(uint8_t chip) {
  wg12232_set_data_port_direction(0x00);
  WG12232_DATA_PORT = 0x00;
  WG12232_A0RW_PORT |= (1 << WG12232_RW_PIN);
  WG12232_A0RW_PORT &= ~(1 << WG12232_A0_PIN);
  wg12232_strobe_on(chip);
  wg12232_delay();
  uint8_t res = wg12232_read_data_port_pins();
  wg12232_strobe_off(chip);
  WG12232_A0RW_PORT &= ~(1 << WG12232_RW_PIN);
  return res;
}

uint8_t wg12232_read(uint8_t chip) {
  wg12232_set_data_port_direction(0x00);
  WG12232_DATA_PORT = 0x00;
  WG12232_A0RW_PORT |= (1 << WG12232_RW_PIN);
  WG12232_A0RW_PORT |= (1 << WG12232_A0_PIN);
  wg12232_strobe_on(chip);
  wg12232_delay();
  asm volatile ("nop");
  uint8_t res = wg12232_read_data_port_pins();
  wg12232_strobe_off(chip);
  WG12232_A0RW_PORT &= ~(1 << WG12232_RW_PIN);
  WG12232_A0RW_PORT &= ~(1 << WG12232_A0_PIN);
  return res;
}

inline void wg12232_select_page(uint8_t page, uint8_t chip) {
  if (((chip & WG12232_CHIP_LEFT) && ((wg12232_chip_page & 0x0F) != page)) || 
      ((chip & WG12232_CHIP_RIGHT) && ((wg12232_chip_page >> 4) != page))) {
    #ifdef WG12232_UPSIDE_DOWN
      wg12232_command(WG12232_CMD_PAGE + 3 - page, chip); 
    #else
      wg12232_command(WG12232_CMD_PAGE + page, chip); 
    #endif
    if (chip & WG12232_CHIP_LEFT)
      wg12232_chip_page = (wg12232_chip_page & 0xF0) | page;
    if (chip & WG12232_CHIP_RIGHT)
      wg12232_chip_page = (wg12232_chip_page & 0x0F) | (page << 4);
  }
}

inline void wg12232_select_col(uint8_t col, uint8_t chip) {
  #ifdef WG12232_UPSIDE_DOWN
  wg12232_command(WG12232_CMD_COLUMN + DISPLAY_START_COLUMN + col, chip);
  #else
  wg12232_command(WG12232_CMD_COLUMN + col, chip);
  #endif
}

void display_init() {
  WG12232_DATA_PORT = 0;
  wg12232_set_data_port_direction(0x00);
  WG12232_A0RW_PORT &= ~((1 << WG12232_A0_PIN) | (1 << WG12232_RW_PIN));
  if (&WG12232_A0RW_PORT == &PORTA) {
    DDRA |= (1 << WG12232_A0_PIN) | (1 << WG12232_RW_PIN);
  } else if (&WG12232_A0RW_PORT == &PORTB) {
    DDRB |= (1 << WG12232_A0_PIN) | (1 << WG12232_RW_PIN);
  } else if (&WG12232_A0RW_PORT == &PORTC) {
    DDRC |= (1 << WG12232_A0_PIN) | (1 << WG12232_RW_PIN);
  } else if (&WG12232_A0RW_PORT == &PORTD) {
    DDRD |= (1 << WG12232_A0_PIN) | (1 << WG12232_RW_PIN);
  }
  WG12232_CS_PORT &= ~((1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN));
  if (&WG12232_CS_PORT == &PORTA) {
    DDRA |= (1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN);
  } else if (&WG12232_CS_PORT == &PORTB) {
    DDRB |= (1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN);
  } else if (&WG12232_CS_PORT == &PORTC) {
    DDRC |= (1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN);
  } else if (&WG12232_CS_PORT == &PORTD) {
    DDRD |= (1 << WG12232_CS1_PIN) | (1 << WG12232_CS2_PIN);
  }

  wg12232_command(WG12232_CMD_RESET, WG12232_CHIP_BOTH); // Сброс Reset
  for (uint8_t i = 255; i != 0; i--) { 
    if (((wg12232_status(WG12232_CHIP_LEFT) & 0x90) == 0) && ((wg12232_status(WG12232_CHIP_RIGHT) & 0x90) == 0))
      break;
    _delay_ms(1);
  }
  wg12232_command(WG12232_CMD_ENABLE, WG12232_CHIP_BOTH); // Включение Turn on
  wg12232_command(WG12232_CMD_START_LINE + 0, WG12232_CHIP_BOTH); // Прокрутка дисплея - 0 Scroll - 0
  wg12232_command(WG12232_CMD_DUTY_CYCLE_32, WG12232_CHIP_BOTH); // широта импульса 1/32 Duty Cycle 1/32
  #ifdef WG12232_UPSIDE_DOWN 
    wg12232_command(WG12232_CMD_ADC_CCW, WG12232_CHIP_BOTH); // Режим CCW (для перевёрнутого дисплея) CCW for upside-down display
  #else
    wg12232_command(WG12232_CMD_ADC_CW, WG12232_CHIP_BOTH); // Режим CW, for normally oriented display
  #endif
  display_clear();
}

void display_fill(uint8_t b) {
  for (uint8_t page = 0; page < DISPLAY_PAGES; page++) {
    wg12232_command(WG12232_CMD_PAGE + page, WG12232_CHIP_BOTH);
    wg12232_command(WG12232_CMD_COLUMN + DISPLAY_START_COLUMN, WG12232_CHIP_BOTH);
    for (uint8_t x = 0; x < DISPLAY_HALF_WIDTH; x++) {
      wg12232_write(b, WG12232_CHIP_BOTH);
    }
  }
  #ifdef WG12232_UPSIDE_DOWN
  wg12232_chip_page = 0x00;
  #else
  wg12232_chip_page = ((DISPLAY_PAGES - 1) << 4) | (DISPLAY_PAGES - 1);
  #endif
  wg12232_select_col(0, WG12232_CHIP_BOTH);
}
 
inline void display_clear() {
  display_fill(0);
}

/*
inline uint8_t reverse_bits(uint8_t x) {
   x = ((x & 0xAA) >>  1) | ((x & 0x55) <<  1);
   x = ((x & 0xCC) >>  2) | ((x & 0x33) <<  2);
   x = (x >> 4) | (x <<  4);
   return x;
}
*/

inline uint8_t _select_chip_and_page(int8_t x, uint8_t page) {
  if (x < DISPLAY_HALF_WIDTH) {
    wg12232_select_page(page, WG12232_CHIP_LEFT);
    wg12232_select_col(x, WG12232_CHIP_LEFT);
    return WG12232_CHIP_LEFT;
  } else {
    wg12232_select_page(page, WG12232_CHIP_RIGHT);
    wg12232_select_col((x - DISPLAY_HALF_WIDTH), WG12232_CHIP_RIGHT);
    return WG12232_CHIP_RIGHT;
  }
}

inline uint8_t _start_right_chip(uint8_t page) {
  wg12232_select_page(page, WG12232_CHIP_RIGHT);
  wg12232_select_col(0, WG12232_CHIP_RIGHT);
  return WG12232_CHIP_RIGHT;
}

void _display_vbits_pgm_masked(int8_t x, uint8_t page, PGM_VOID_P data, int8_t shft, uint8_t count, uint8_t andMask, uint8_t xorMask, uint8_t msbUp) {
  uint8_t dummyRead = 0;
  uint8_t chip = _select_chip_and_page(x, page);
  do {
    if (chip == WG12232_CHIP_LEFT) {
      if (x >= DISPLAY_HALF_WIDTH) {
        chip = _start_right_chip(page);
        dummyRead = 0;
      }
    } else {
      if (x >= DISPLAY_WIDTH) {
        return;
      }
    }
    uint8_t byteToWrite = pgm_read_byte(data++);
    #ifdef WG12232_UPSIDE_DOWN
    if (!msbUp)
      byteToWrite = reversebits(byteToWrite);
    #else
    if (msbUp)
      byteToWrite = reversebits(byteToWrite);
    #endif
    if (shft != 0) {
      byteToWrite = (shft < 0) ? shl(byteToWrite, -shft) : shr(byteToWrite, shft);
    }
    if (!dummyRead) {
      wg12232_read(chip);
      dummyRead = 0xFF;
    }
    uint8_t screenData = wg12232_read(chip);
    uint8_t newData = ((screenData & andMask) | byteToWrite) ^ xorMask;
    if (newData != screenData) {
      wg12232_select_col((chip == WG12232_CHIP_LEFT) ? x : (x - DISPLAY_HALF_WIDTH), chip);
      wg12232_write(newData, chip);
      dummyRead = 0;
    }
    count--;
    x++;
  } while (count > 0);
}

void display_vbits_pgm(int8_t x, int8_t y, PGM_VOID_P data, uint8_t count, uint8_t andMask, uint8_t xorMask, uint8_t msbUp) {
  if ((y <= -8) || (y >= DISPLAY_HEIGHT) || (x + count <= 0) || (x >= DISPLAY_WIDTH)) 
    return;
  if (x < 0) {
    data -= x;
    count += x;
    x = 0;
  }
  if (count == 0)
    return;
  #ifdef WG12232_UPSIDE_DOWN
  if (!msbUp) {
  #else
  if (msbUp) {
  #endif
    if ((xorMask != 0) && (xorMask != 0xFF))
      xorMask = reversebits(xorMask);
    if ((andMask != 0) && (andMask != 0xFF))
      andMask = reversebits(andMask);
  }

  uint8_t page = y >> 3;
  if ((y & 7) == 0) {  // Если попали целиком в нужный байт
    if (andMask == 0x00) { // Если полностью затираем изображение на экране
      uint8_t chip = _select_chip_and_page(x, page);
      do {
        if (chip == WG12232_CHIP_LEFT) {
          if (x >= DISPLAY_HALF_WIDTH) {
            chip = _start_right_chip(page);
          } 
        } else {
          if (x >= DISPLAY_WIDTH) {
            return;
          } 
        }
        uint8_t byteToWrite = pgm_read_byte(data++);
        #ifdef WG12232_UPSIDE_DOWN 
        if (!msbUp) 
          byteToWrite = reversebits(byteToWrite);
        #else
        if (msbUp) 
          byteToWrite = reversebits(byteToWrite);
        #endif
        wg12232_write(byteToWrite ^ xorMask, chip);
        x++;
        count--;
      } while (count > 0);
    } else { // Накладываем маску
      _display_vbits_pgm_masked(x, page, data, 0,count, andMask, xorMask, msbUp);
    }
  } else { // Если выводимая информация затрагивает две страницы
    if (page < DISPLAY_PAGES) {
      uint8_t sh = (y & 7);
      #ifdef WG12232_UPSIDE_DOWN
      _display_vbits_pgm_masked(x, page, data, sh, count, ~shr(~andMask, sh), shr(xorMask, sh), msbUp);
      #else
      _display_vbits_pgm_masked(x, page, data, -sh, count, ~shl(~andMask, sh), shl(xorMask, sh), msbUp);
      #endif
    }
    page = (y + 8) >> 3;
    if (page < DISPLAY_PAGES) {
      uint8_t sh = 8 - (y & 7);
      #ifdef WG12232_UPSIDE_DOWN
      _display_vbits_pgm_masked(x, page, data, -sh, count, ~shl(~andMask, sh), shl(xorMask, sh), msbUp);
      #else
      _display_vbits_pgm_masked(x, page, data, sh, count, ~shr(~andMask, sh), shr(xorMask, sh), msbUp);
      #endif
    }
  }
  
}

void _display_modify_vbits(int8_t x, uint8_t page, uint8_t andMask, uint8_t xorMask) {
  uint8_t chip = _select_chip_and_page(x, page);
  if (andMask == 0x00) { // Если полностью затираем изображение на экране
    wg12232_write(xorMask, chip);
  } else {
    wg12232_read(chip); // холостое чтение
    uint8_t screenData = wg12232_read(chip);
    uint8_t newData = (screenData & andMask) ^ xorMask;
    if (newData != screenData) {
      wg12232_select_col((chip == WG12232_CHIP_LEFT) ? x : x - DISPLAY_HALF_WIDTH, chip);
      wg12232_write(newData, chip);
    }
  }

}
void display_vbits_single(int8_t x, int8_t y, uint8_t andMask, uint8_t xorMask, uint8_t msbUp) {
  if ((y <= -8) || (y >= DISPLAY_HEIGHT) || (x < 0) || (x >= DISPLAY_WIDTH))
    return;
  #ifdef WG12232_UPSIDE_DOWN
  if (!msbUp) {
  #else
  if (msbUp) {
  #endif
    if ((xorMask != 0) && (xorMask != 0xFF))
      xorMask = reversebits(xorMask);
    if ((andMask != 0) && (andMask != 0xFF))
      andMask = reversebits(andMask);
  }
  uint8_t page = y >> 3;
  if ((y & 7) == 0) {  // Если попали целиком в нужный байт
    _display_modify_vbits(x, page, andMask, xorMask);
  } else {
    if (page < DISPLAY_PAGES) {
      uint8_t sh = (y & 7);
      #ifdef WG12232_UPSIDE_DOWN
       _display_modify_vbits(x, page, ~shr(~andMask, sh), shr(xorMask, sh));
      #else
       _display_modify_vbits(x, page, ~shl(~andMask, sh), shl(xorMask, sh));
      #endif
    }
    page = (y + 8) >> 3;
    if (page < DISPLAY_PAGES) {
      uint8_t sh = 8 - (y & 7);
      #ifdef WG12232_UPSIDE_DOWN
      _display_modify_vbits(x, page, ~shl(~andMask, sh), shl(xorMask, sh));
      #else
      _display_modify_vbits(x, page, ~shr(~andMask, sh), shr(xorMask, sh));
      #endif
    }
  }
}


void _display_fill_bits_modify(uint8_t x, uint8_t count, uint8_t page, uint8_t andMask, uint8_t xorMask, uint8_t andInv, uint8_t xorInv) {
  uint8_t chip = _select_chip_and_page(x, page);
  wg12232_command(WG12232_CMD_RMW_START, chip);
  do {
    if ((chip == WG12232_CHIP_LEFT) && (x >= DISPLAY_HALF_WIDTH)) {
      wg12232_command(WG12232_CMD_RMW_END, WG12232_CHIP_LEFT);
      chip = _start_right_chip(page);
      wg12232_command(WG12232_CMD_RMW_START, chip);
    }
    wg12232_read(chip); // dummy
    uint8_t screenData = wg12232_read(chip);
    wg12232_write((screenData & andMask) ^ xorMask , chip);
    xorMask ^= xorInv;
    andMask ^= andInv;
    x++;
    count--;
  } while (count > 0);
  wg12232_command(WG12232_CMD_RMW_END, chip);
}

void display_fill_rect(int8_t left, int8_t top, int8_t width, int8_t height, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd) {
  if ((width < 0) || (height < 0))
    return;
  if (left < 0) {
    width += left;
    left = 0;
  }
  if (top < 0) {
    height += top;
    top = 0;
  }
  if (width > DISPLAY_WIDTH - left) {
    width = DISPLAY_WIDTH - left;
  }
  if (height > DISPLAY_HEIGHT - top) {
    height = DISPLAY_HEIGHT - top;
  }
  if ((width < 0) || (height < 0))
    return;
  xorOdd ^= xorEven;
  andOdd ^= andEven;
  if (left & 1) {
    xorEven ^= xorOdd;
    andEven ^= andOdd;
  }

  uint8_t page = top >> 3;
  if ((top & 7) != 0) {
    uint8_t sh = (top & 7);
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t mask = shr(0xFF, sh);
    #else
    uint8_t mask = shl(0xFF, sh);
    #endif

    if ((sh + height) < 8) {
      uint8_t sh2 = 8 - sh - height;
      #ifdef WG12232_UPSIDE_DOWN
      mask &= shl(0xFF, sh2);
      #else
      mask &= shr(0xFF, sh2);
      #endif
    }
    _display_fill_bits_modify(left, width, page, andEven | ~mask, xorEven & mask, andOdd & mask, xorOdd & mask);
    if ((sh + height) <= 8)
      return;
    height -= (8 - sh);
    page++;
  }
  while (height >= 8) {
    if ((andEven | andOdd) == 0) { // Полностью затирается экран
      uint8_t xor = xorEven;
      uint8_t x = left;
      uint8_t chip = _select_chip_and_page(x, page);
      uint8_t count = width;
      do {
        if ((chip == WG12232_CHIP_LEFT) && (x >= DISPLAY_HALF_WIDTH)) 
          chip = _start_right_chip(page);
        wg12232_write(xor, chip);
        xor ^= xorOdd;
        x++;
        count--;
      } while (count > 0);
    } else {
      _display_fill_bits_modify(left, width, page, andEven, xorEven, andOdd, xorOdd);
    }
    height -= 8;
    page++;
  } 
  if (height > 0) {
    uint8_t sh = 8 - height;
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t mask = 0xFF << sh;
    #else
    uint8_t mask = shr(0xFF, sh);
    #endif
    _display_fill_bits_modify(left, width, page, andEven | ~mask, xorEven & mask, andOdd & mask, xorOdd & mask);
  }
}
 
void display_hline(int8_t x1, int8_t x2, int8_t y, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd) {
  if ((y < 0) || (y >= DISPLAY_HEIGHT)) 
    return;
  if (x1 > x2) {
    int8_t t = x1;
    x1 = x2;
    x2 = t;
  }
  if (x1 < 0)
    x1 = 0;
  if (x2 >= DISPLAY_WIDTH)
    x2 = (DISPLAY_WIDTH - 1);
  if (x2 < x1)
    return;
  andOdd ^= andEven;
  xorOdd ^= xorEven;
  if (x1 & 1) {
    xorEven ^= xorOdd;
    andEven ^= andOdd;
  }
  uint8_t page = y >> 3;
  uint8_t bit = (y & 7);
  #ifdef WG12232_UPSIDE_DOWN
    uint8_t mask = shr(0x80, bit);
  #else 
    uint8_t mask = shl(1, bit);
  #endif
  _display_fill_bits_modify(x1, x2 - x1 + 1, page, andEven | ~mask, xorEven & mask, andOdd & mask, xorOdd & mask);
}

void display_vline(int8_t x, int8_t y1, int8_t y2, uint8_t and, uint8_t xor) {
  if ((x < 0) || (x >= DISPLAY_WIDTH))
    return;
  if (y1 > y2) {
    int8_t t = y1;
    y1 = y2;
    y2 = t;
  }
  if (y1 < 0)
    y1 = 0;
  if (y2 >= DISPLAY_HEIGHT)
    y2 = (DISPLAY_HEIGHT - 1);
  if (y2 < y1)
    return;
  y2 -= y1 + 1;
  uint8_t page = y1 >> 3;
  if ((y1 & 7) != 0) {
    uint8_t sh = (y1 & 7);
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t mask = shr(0xFF, sh);
    #else
    uint8_t mask = shl(0xFF, sh);
    #endif

    if ((sh + y2) < 8) {
      uint8_t sh2 = 8 - sh - y2;
      #ifdef WG12232_UPSIDE_DOWN
      mask &= shl(0xFF, sh2);
      #else
      mask &= shr(0xFF, sh2);
      #endif
    }
    _display_modify_vbits(x, page, and | ~mask, xor & mask);
    if ((sh + y2) <= 8)
      return;
    y2 -= (8 - sh);
    page++;
  }
  while (y2 >= 8) {
    _display_modify_vbits(x, page, and, xor);
    y2 -= 8;
    page++;
  }
  if (y2 > 0) {
    uint8_t sh = 8 - y2;
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t mask = shl(0xFF, sh);
    #else
    uint8_t mask = shr(0xFF, sh);
    #endif
    _display_modify_vbits(x, page, and | ~mask, xor & mask);
  }
}

void display_pixel(int8_t x, int8_t y, uint8_t and, uint8_t xor) {
  if ((x < 0) || (x >= DISPLAY_WIDTH) || (y < 0) || (y >= DISPLAY_HEIGHT))
    return;
  uint8_t page = y >> 3;
  uint8_t sh = (y & 7);
  #ifdef WG12232_UPSIDE_DOWN
  uint8_t mask = shr(0x80, sh);
  #else
  uint8_t mask = shl(1, sh);
  #endif
  _display_modify_vbits(x, page, and | ~mask, xor & mask);
}

void display_line(int8_t x1, int8_t y1, int8_t x2, int8_t y2, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd) {
  if (x1 > x2) {
    uint8_t t;
    t = x1; x1 = x2; x2 = t;
    t = y1; y1 = y2; y2 = t;
  }
  if ((x2 < 0) || (x1 >= DISPLAY_WIDTH))
    return;
  int16_t dy = y2 -y1;
  if (dy == 0) {
    display_hline(x1, x2, y1, andEven, xorEven, andOdd, xorOdd);
    return;
  }
  uint8_t dx = x2 - x1;
  if (dx == 0) {
    display_vline(x1, y1, y2, (x1 & 1) ? andOdd : andEven, (x1 & 1) ? xorOdd : xorEven);
    return;
  }
  int8_t dydir;
  uint8_t dyabs;
  if (dy > 0) {
    dydir = 1;  
    dyabs = dy;
  } else {
    dydir = -1;
    dyabs = -dy;
  }
  if (x2 >= DISPLAY_WIDTH)
    x2 = DISPLAY_WIDTH - 1;
  xorOdd ^= xorEven;
  andOdd ^= andEven;
  if (dyabs > dx) { // Шагаем по вертикали быстрее, чем по горизонтали
    uint8_t offset = dyabs >> 1;
    // Сначала довести линию до зоны рисования, если она находится вне неё
    while (x1 < 0) {
      if (dx >= (dyabs - offset)) { offset += dx - dyabs; x1++; } else { offset += dx; }
      y1 += dydir;
    }
    if (y1 < 0) {
      if (dydir < 0)
        return;
      do {
        if (y1 == y2) return;
        if (dx >= (dyabs - offset)) { offset += dx - dyabs; x1++; } else { offset += dx; }
        y1 += dydir;
      } while (y1 < 0);
    } else if (y1 >= DISPLAY_HEIGHT) {
      if (dydir > 0)
      return;
      do {
        if (y1 == y2) return;
        if (dx >= (dyabs - offset)) { offset += dx - dyabs; x1++; } else { offset += dx; }
        y1 += dydir;
      } while (y1 >= DISPLAY_HEIGHT);
    }
    if (x1 > x2)
      return;
    uint8_t page = ((uint8_t)y1) >> 3;
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t ymask = shr(0x80, ((uint8_t)y1) & 7);
    #else
    uint8_t ymask = shl(1, ((uint8_t)y1) & 7);
    #endif
    if (x1 & 1) {
      xorEven ^= xorOdd;
      andEven ^= andOdd;
    }
    uint8_t chip = _select_chip_and_page(x1, page);
    wg12232_read(chip); // dummy;
    uint8_t screenData = wg12232_read(chip);
    uint8_t data = screenData;
    for(;;) {
      data = (data & (andEven | ~ymask)) ^ (xorEven & ymask);
      if (y1 == y2) // Последний пиксель
        break;
      y1 += dydir;
      if ((y1 >= DISPLAY_HEIGHT) || (y1 < 0))
        break;

      uint8_t pageChange = 0;
      #ifdef WG12232_UPSIDE_DOWN
      if (dydir >= 0)
      #else
      if (dydir < 0)
      #endif
        ymask >>= 1;
      else
        ymask <<= 1;
      if (ymask == 0) { // Если меняется и страница тоже
        page += dydir;
        pageChange = 0xFF;
        #ifdef WG12232_UPSIDE_DOWN
        if (dydir >= 0)
        #else
        if (dydir < 0)
        #endif
          ymask = 0x80;
        else
          ymask = 1;
      }
      if (dx >= (dyabs - offset)) {
        offset += dx - dyabs;

        // Дошли до смены икса
        if (data != screenData) { // Вернёмся и перезапишем
          wg12232_select_col((chip == WG12232_CHIP_LEFT) ? x1 : (x1 - DISPLAY_HALF_WIDTH), chip);
          wg12232_write(data, chip);
          x1++; // Рассмотреть случаи, когда доходим до границы чипа
          if (((chip == WG12232_CHIP_LEFT) && (x1 >= DISPLAY_HALF_WIDTH)) || pageChange) {
             chip = _select_chip_and_page(x1, page);
             pageChange = 0x00;
          }
          wg12232_read(chip); // dummy read
        } else {
          x1++;
          xorEven ^= xorOdd;
          andEven ^= andOdd;
          if (((chip == WG12232_CHIP_LEFT) && (x1 >= DISPLAY_HALF_WIDTH)) || pageChange) {
            chip = _select_chip_and_page(x1, page);
            pageChange = 0x00;
            wg12232_read(chip); // dummy read
          }
        }
        screenData = wg12232_read(chip);
        data = screenData;
      } else {
        offset += dx;
        if (pageChange) {
          if (data != screenData) { // Вернёмся и перезапишем
            wg12232_select_col((chip == WG12232_CHIP_LEFT) ? x1 : (x1 - DISPLAY_HALF_WIDTH), chip);
            wg12232_write(data, chip);
          }
          chip = _select_chip_and_page(x1, page);
          wg12232_read(chip); // dummy read
          screenData = wg12232_read(chip);
          data = screenData;
          pageChange = 0x00;
        }
      }
    }
    if (data != screenData) {
      wg12232_select_col((chip == WG12232_CHIP_LEFT) ? x1 : (x1 - DISPLAY_HALF_WIDTH), chip);
      wg12232_write(data, chip);
    }
  } else { // Шагаем по горизонтали быстрее чем, или так же как по вертикали
    uint8_t offset = dx >> 1;
    // Сначала довести линию до зоны рисования, если она находится вне неё
    while (x1 < 0) {
      if (dyabs >= (dx - offset)) { offset += dyabs - dx; y1 += dydir; } else { offset += dyabs; }
      x1++;
    }
    if (y1 < 0) {
      if (dydir < 0)
        return;
      do {
        if (x1 >= x2) return;
        if (dyabs >= (dx - offset)) { offset += dyabs - dx; y1 += dydir; } else { offset += dyabs; }
        x1++;
      } while (y1 < 0);
    } else if (y1 >= DISPLAY_HEIGHT) {
      if (dydir > 0)
        return;
      do {
        if (x1 >= x2) return;
        if (dyabs >= (dx - offset)) { offset += dyabs - dx; y1 += dydir; } else { offset += dyabs; }
        x1++;
      } while (y1 >= DISPLAY_HEIGHT);
    }
    uint8_t page = ((uint8_t)y1) >> 3;
    #ifdef WG12232_UPSIDE_DOWN
    uint8_t ymask = shr(0x80, ((uint8_t)y1) & 7);
    #else
    uint8_t ymask = shl(1, ((uint8_t)y1) & 7);
    #endif
    if (x1 & 1) {
      xorEven ^= xorOdd;
      andEven ^= andOdd;
    }
    uint8_t chip = _select_chip_and_page(x1, page);
    wg12232_command(WG12232_CMD_RMW_START, chip);
    for(;;) {
      wg12232_read(chip); // dummy
      uint8_t data = wg12232_read(chip); 
      data = (data & (andEven | ~ymask)) ^ (xorEven & ymask);
      wg12232_write(data, chip);
      xorEven ^= xorOdd;
      andEven ^= andOdd;
      if (x1 >= x2)
        break;
      x1++;
      if (dyabs >= (dx - offset)) { 
        offset += dyabs - dx; 
        y1 += dydir; 
        if ((y1 < 0) || (y1 >= DISPLAY_HEIGHT))
          break;
        #ifdef WG12232_UPSIDE_DOWN
        if (dydir >= 0)
        #else
        if (dydir < 0)
        #endif
          ymask >>= 1;
        else 
          ymask <<= 1;
        if (ymask == 0) {
          wg12232_command(WG12232_CMD_RMW_END, chip);
          page += dydir;
          #ifdef WG12232_UPSIDE_DOWN
          if (dydir >= 0)
          #else
          if (dydir < 0)
          #endif
            ymask = 0x80;
          else
            ymask = 1;
          chip = _select_chip_and_page(x1, page);
          wg12232_command(WG12232_CMD_RMW_START, chip);
        }
      } else { 
        offset += dyabs;
        if ((chip == WG12232_CHIP_LEFT) && (x1 >= DISPLAY_HALF_WIDTH)) {
          wg12232_command(WG12232_CMD_RMW_END, chip);
          chip = _start_right_chip(page);
          wg12232_command(WG12232_CMD_RMW_START, chip);
        }
      }
    }
    wg12232_command(WG12232_CMD_RMW_END, chip);
  }

}