﻿/*
 * wg12232.h
 * Модуль для работы с дисплеем WG12232A, или совместимым
 *
 * Author: Погребняк Дмитрий, г. Самара, 2013
 */ 


#ifndef WG12232_H_
#define WG12232_H_

#ifndef F_CPU 
  #define F_CPU 8000000UL
#endif

// Определение WG12232_UPSIDE_DOWN включает режим обратной адресации и переворачивает выводимое на дисплей изображение
// Defining WG12232_UPSIDE_DOWN will turn on CCW-mode and turn image upside-down
//#define WG12232_UPSIDE_DOWN

#define WG12232_DATA_PORT PORTB
#define WG12232_A0RW_PORT PORTA
#define WG12232_A0_PIN 0
#define WG12232_RW_PIN 1
#define WG12232_CS_PORT PORTA
#define WG12232_CS1_PIN 2
#define WG12232_CS2_PIN 3

#include <avr/io.h>
#include <avr/pgmspace.h>

#define WG12232_CMD_ENABLE 0xAF
#define WG12232_CMD_DISABLE 0xAE
#define WG12232_CMD_START_LINE 0xC0
#define WG12232_CMD_PAGE 0xB8
#define WG12232_CMD_COLUMN 0x00
#define WG12232_CMD_ADC_CW 0xA0
#define WG12232_CMD_ADC_CCW 0xA1
#define WG12232_CMD_DUTY_CYCLE_32 0xA9
#define WG12232_CMD_RMW_START 0xE0
#define WG12232_CMD_RMW_END 0xEE
#define WG12232_CMD_RESET 0xE2

#define DISPLAY_WIDTH 122
#define DISPLAY_HEIGHT 32
#define DISPLAY_HALF_WIDTH (DISPLAY_WIDTH / 2)
#define DISPLAY_PAGES (DISPLAY_HEIGHT / 8)

#ifdef WG12232_UPSIDE_DOWN 
  #define DISPLAY_START_COLUMN 19
  #define WG12232_CHIP_LEFT 2
  #define WG12232_CHIP_RIGHT 1
#else
  #define DISPLAY_START_COLUMN 0
  #define WG12232_CHIP_LEFT 1
  #define WG12232_CHIP_RIGHT 2
#endif
#define WG12232_CHIP_BOTH 3

void wg12232_command(uint8_t command, uint8_t chip);

void wg12232_write(uint8_t data, uint8_t chip);

uint8_t wg12232_status(uint8_t chip);

uint8_t wg12232_read(uint8_t chip);

void display_vbits_pgm(int8_t x, int8_t y, PGM_VOID_P data, uint8_t count, uint8_t andMask, uint8_t xorMask, uint8_t msbUp);

void display_vbits_single(int8_t x, int8_t y, uint8_t andMask, uint8_t xorMask, uint8_t msbUp);

/*
  Конфигурирует порты, сбрасывает, инициализирует и  очищает дисплей
  Perform the configuration of port pins, reset, initialize and clear wg12232a display 
*/
void display_init();

/*
  Заполняет дисплей указанным байтом
  Fill the display area with specified byte
*/
void display_fill(uint8_t b);

/*
  Очищает экран
  Clear screen
*/
void display_clear();

/*
  Закрашивает прямоугольник изменяя изображение на дисплее по формуле:
  d = (d & andXxxx) ^ xorXxxx, где Xxxx - Even - для чётных горизонтальных позиций, Odd - для нечётных
  Fill the rectangle
  Modify image data using formula: d = (d & andXxxx) ^ xorXxxx, where Xxxx - Even or Odd, depended on in which column to be modified

  left - левая координата
  top - верхняя координата
  width - ширина
  height - высота
  andEven, xorEven, andOdd, xorOdd
  0, 0, 0, 0 - очистить дисплей/clear
  0, 0xFF, 0, 0xFF - закрасить дисплей/fill
  0xFF, 0xFF, 0xFF, 0xFF - инвертировать/invert
  0, 0xAA, 0, 0x55 - закрасить точки в шахматном порядке/chessfill
  0x55, 0xAA, 0xAA, 0x55 - эффект затемнения/shade effect
*/
void display_fill_rect(int8_t left, int8_t top, int8_t width, int8_t height, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd);

/*
  Рисует горизонтальную линию.
  Draw horizontal line
  x1 - координата начала / start of line
  x2 - координата конца (включительно) / end of line (inclusive)
  y - вертикальная координата / vertical position
  andEven, xorEven, andOdd, xorOdd - (см. display_fill_rect)
  0, 0xFF, 0, 0xFF - нарисовать линию, draw line
  0, 0, 0, 0 - стереть линию, clear line
  0x55, 0xAA, 0xAA, 0x55 - нарисовать линию через точку, dot line
*/
void display_hline(int8_t x1, int8_t x2, int8_t y, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd);

/*
  Рисует вертикальную линию.
  Draw vertical line
  x - горизонтальная координата / horizontal position
  y1 - координата начала / start of line
  y2 - координата конца (включительно) / end of line (inclusive)
  andEven, xorEven, andOdd, xorOdd - (см. display_fill_rect)
  0, 0xFF, 0, 0xFF - нарисовать линию, draw line
  0, 0, 0, 0 - стереть линию, clear line
  0x55, 0xAA, 0xAA, 0x55 - нарисовать линию через точку, dot line
*/
void display_vline(int8_t x, int8_t y1, int8_t y2, uint8_t and, uint8_t xor);

/*
  Рисует точку.
  Draw the pixel
  x - горизонтальная координата / horizontal position
  y - вертикальная координата / vertical position
  and, xor - (см. display_fill_rect)
  0, 0xFF - закрасить пиксель, set pixel
  0, 0 - стереть пиксель, clear pixel
  0xFF, 0xFF - инвертировать пиксель, invert pixel
*/
void display_pixel(int8_t x, int8_t y, uint8_t and, uint8_t xor);

/*
  Рисует вертикальную линию.
  Draw vertical line
  x1, y1 - координаты начала / start of line
  x2, y2 - координаты конца (включительно) / end of line (inclusive)
  andEven, xorEven, andOdd, xorOdd - (см. display_fill_rect)
  0, 0xFF, 0, 0xFF - нарисовать линию, draw line
  0, 0, 0, 0 - стереть линию, clear line
  0x55, 0xAA, 0xAA, 0x55 - нарисовать линию через точку, dot line
*/
void display_line(int8_t x1, int8_t y1, int8_t x2, int8_t y2, uint8_t andEven, uint8_t xorEven, uint8_t andOdd, uint8_t xorOdd);


#endif /* WG12232_H_ */